# $Name:  $ 
# $Author: dgp $ 
# $Date: 1997/03/19 21:13:54 $ 
#
# A library of Tcl procs which provide conversions between different 
# enumeration data types.  
#
# New procs are added to this library as they are needed
# by other Tcl libraries, so it may not be complete.

# Add Enumeration package to database; define standard global variables
OMFCVS_PackageProvide {^Name: Enumeration-0_0a0^} enum

# proc RomanDigitMap digit symbols
#	Convert a decimal digit to a Roman numeral representation,
#	using the symbols in the 3-character string symbols as
#	the 10, 5, and 1 representer.
proc _OMFENUM_RomanDigitMap {digit symbols} {
  set i [string index $symbols 2]
  set v [string index $symbols 1]
  set x [string index $symbols 0]
  switch $digit {
    0	{return ""}
    1	{return $i}
    2	{return $i$i}
    3	{return $i$i$i}
    4	{return $i$v}
    5	{return $v}
    6	{return $v$i}
    7	{return $v$i$i}
    8	{return $v$i$i$i}
    9	{return $i$x}
  }
}

# proc NaturalToRoman value
#	Convert a natural number to its Roman numeral representation.
proc OMFENUM_NaturalToRoman {value} {
  if {![regexp {[1-9][0-9]*} $value]} {
    return -code error \
      -errorinfo "expected positive integer but got \"$value\"" ""
  }
  if {$value > 3999} {
    return -code error \
      -errorinfo "\"$value\" exceeds range translatable to Roman numerals" ""
  }
  set result ""
  # Handle any thousands
  if {[string length $value] == 4} {
    for {set i 0} {$i < [string index $value 0]} {incr i} {
      append result m
    }
  }
  set elements "mdclxvi"
  # Handle the rest
  for {set digits 3} {$digits > 0} {incr digits -1} {
    if {$digits == [string length $value]} {
      append result [_OMFENUM_RomanDigitMap [string index $value 0] \
			[string range $elements [expr 6 - 2 * $digits] \
			[expr 8 - 2 * $digits] ] ]
      regsub {^[0-9]} $value {} value
    }
  }
  return $result
}

# proc OMFENUM_NaturalToAlpha value
#	Convert a positive number value to an alphabetic representation.
#	Use a-z as "digits" of base 26 system, sort of.
#	Assumes a-z contiguous in character set, like ASCII
proc OMFENUM_NaturalToAlpha {value} {
  if {![regexp {[1-9][0-9]*} $value]} {
    return -code error \
           -errorinfo "expected positive integer but got \"$value\"" \
           ""
  }
  set result {}
  incr value -1
  if {$value > 25} {
    set l [expr $value / 26]
    set result [toalpha $l]
    set value [expr $value % 26]
  }
  append result [format %c [expr $value + \
		[set v 0; scan a %c v; set v]]]
  return $result
}

