# FILE: dialog.tk
#
# Tcl/Tk scripts for various dialog boxes
#
# Requires Tcl/Tk script file: misc.tk
#  It must be source'd _before_ this script.
#

puts "dialog.tcl ..."

###################### OMF_PrintDialog ###################################
# 4-Feb-1997, mjd
#
# "Local" global variables are stored in array OMFpd$id(), where id is
# obtained by converting .'s to _'s in the $window value passed
# to OMF_PrintDialog

proc OMF_PrintDialogKill { window } {
    OMF_SetFocus . $window ;# Transfer focus back to root window
                            # before destroying current window
    OMF_GetID $window id base
    destroy $base   ;# Important: You will be unable to delete
                     # $printvar until _after_ $base (or at least
                     # the entry widget tied to $printvar) is
                     # destroyed!
    set printvar "OMFpd$id"
    global $printvar
    trace vdelete $printvar w OMF_PrintDialogTrace
    global UpdateViewCount
    trace vdelete UpdateViewCount w OMF_PrintDialogViewportTrace$id
    unset $printvar
}

# Determines width:height ratio using values of 
# pvar(canvas) and pvar(crop)
proc OMF_PrintDialogSetSizeRatio { pvarname } {
    upvar #0 $pvarname pvar
    if {$pvar(crop)} {
        # Printing cropped (viewable) region
        set width [$pvar(canvas) cget -width]
        set height [$pvar(canvas) cget -height]
    } else {
        # Printing whole canvas
        set sr [.canvas_frame.canvas cget -scrollregion]
        set width [expr [lindex $sr 2] - [lindex $sr 0]]
        set height [expr [lindex $sr 3] - [lindex $sr 1]]
    }
    if {$width<=0}  { set width 1. }   ;# Safety: Both ratio &
    if {$height<=0} { set height 1. }  ;# 1/ratio should be valid.
    global $pvarname
    set ${pvarname}(sizeratio) [expr double($width)/double($height)]
    # We need to access pvar(sizeratio) as ${pvarname}(sizeratio)
    # in order to get the trace on ${pvarname} to work properly.
}

# Initialize OMF_PrintDialog data array from outside print array
proc OMF_PrintDialogCopyIn { window parrname } {
    OMF_GetID $window id
    set pvarname "OMFpd$id"
    upvar #0 $pvarname pvar
    upvar #0 $parrname parr
    set pvar(window)    $window
    set pvar(canvas)    $parr(canvas)
    set pvar(filename)  $parr(filename)
    set pvar(orient)    $parr(orient)
    set pvar(papertype) $parr(papertype)
    set pvar(crop)      $parr(crop)
    # Position is broken into two variables inside OMF_PrintDialog
    set pvar(verpos) {} ;# Default is centered
    regexp -nocase {[ns]} $parr(position) pvar(verpos)
    set pvar(horpos) {} ;# Default is centered
    regexp -nocase {[ew]} $parr(position) pvar(horpos)
    # Size is handled differently internal to OMF_PrintDialog:
    set pvar(units) "i"  ;# Safety
    regexp -nocase {[ic]} $parr(xmargin) pvar(units)
    set pvar(oldunits) $pvar(units) ;# Used to track unit conversions
    set pvar(leftmargin) [OMF_ConvertUnits $parr(xmargin) $pvar(units)]
    set pvar(topmargin) [OMF_ConvertUnits $parr(ymargin) $pvar(units)]
    OMF_PrintDialogSetSizeRatio $pvarname
    set direction width; set length 1i  ;# Safety
    if ![regexp -- {-page(width|height) +([0-9.]+.)} $parr(size) match \
            direction length] {
        OMF_FatalError "Invalid print size switch ($parr(size)) detected" \
                "OMF_PrintDialogCopyIn Error"
    }
    if {"$direction"=="width"} {
        set width  [OMF_ConvertUnits $length $pvar(units)]
        set height [expr $width/$pvar(sizeratio)]
    } else {
        set height [OMF_ConvertUnits $length $pvar(units)]
        set width  [expr $height*$pvar(sizeratio)]
    }

    # Strip extra precision off of width & height
    set width [string trim $width]
    if {[string compare $width "."]==0} {
        set width "0."
    }
    if [regexp {\....} $width] {
        set width [format "%.2f" $width]
    }

    set height [string trim $height]
    if {[string compare $height "."]==0} {
        set height "0."
    }
    if [regexp {\....} $height] {
        set height [format "%.2f" $height]
    }

    set pvar(width)  $width
    set pvar(height) $height
}

# Copy OMF_PrintDialog data array to outside print array
proc OMF_PrintDialogCopyOut { window parrname } {
    OMF_GetID $window id
    set printvar "OMFpd$id"
    upvar #0 $printvar pvar
    upvar #0 $parrname parr
    set parr(filename)  $pvar(filename)
    set parr(orient)    $pvar(orient)
    set parr(papertype) $pvar(papertype)
    set parr(crop)      $pvar(crop)
    # Position is broken into two variables inside OMF_PrintDialog
    set parr(position)  "$pvar(verpos)$pvar(horpos)"
    if {"$parr(position)"==""} { set parr(position) "c" }

    # Size is handled differently internal to OMF_PrintDialog:
    set width [string trim $pvar(width)]
    if {[string compare $width ""]==0 \
        || [string compare $width "."]==0} {
        set width 0.
    }
    set parr(size)  "-pagewidth $width$pvar(units)"

    set leftmargin [string trim $pvar(leftmargin)]
    if {[string compare $leftmargin ""]==0 \
        || [string compare $leftmargin "."]==0} {
        set leftmargin 0.
    }
    set parr(xmargin) "$leftmargin$pvar(units)"

    set topmargin [string trim $pvar(topmargin)]
    if {[string compare $topmargin ""]==0 \
        || [string compare $topmargin "."]==0} {
        set topmargin "0."
    }
    set parr(ymargin) "$topmargin$pvar(units)"
}

# Fix up variables that depend on viewport display
proc OMF_PrintDialogUpdateVariables { pvarname } {
    OMF_PrintDialogSetSizeRatio $pvarname
}

# Trace procedure to automatically update PrintDialog variables
proc OMF_PrintDialogTrace { name elt op } {
    global $name
    if ![info exists ${name}($elt)] return
    # Disable tracing because we will be modifying several
    # elements of pvar.  Note: We must use the name "pvar"
    # inside the 'trace vdelete' command instead of "$name"
    # because that is how it is accessed inside this routine.
    upvar #0 $name pvar  ;# The #0 is needed for values
                          # modified inside entry widgets.
    trace vdelete pvar w OMF_PrintDialogTrace
    switch -exact -- $elt {
        sizeratio -
        width   {
            set temp [OMF_MakeValidPosNumber $pvar(width)]
            set pvar(width) $temp
            if {[string compare $temp ""]==0 \
                    || [string compare $temp "."]==0 } {
                set pvar(height) ""
            } else {
                set pvar(height) [expr $temp/$pvar(sizeratio)]
            }
        }
        height  {
            set temp [OMF_MakeValidPosNumber $pvar(height)]
            set pvar(height) $temp
            if {[string compare $temp ""]==0 \
                    || [string compare $temp "."]==0 } {
                set pvar(width) ""
            } else {
                set pvar(width) [expr $temp*$pvar(sizeratio)]
            }
        }
        crop    {
            OMF_PrintDialogSetSizeRatio $name
            set temp [OMF_MakeValidPosNumber $pvar(width)]
            set pvar(width) $temp
            if {[string compare $temp ""]==0 \
                    || [string compare $temp "."]==0 } {
                set pvar(height) ""
            } else {
                set pvar(height) [expr $temp/$pvar(sizeratio)]
            }
        }
        units {
            set pvar(units) [string trim $pvar(units)] ;# Safety
            if {[string compare $pvar(units) $pvar(oldunits)]!=0} {
                set pvar(leftmargin) [OMF_ConvertUnits \
                        "$pvar(leftmargin)$pvar(oldunits)" $pvar(units)]
                set pvar(topmargin) [OMF_ConvertUnits \
                        "$pvar(topmargin)$pvar(oldunits)" $pvar(units)]
                set pvar(width) [OMF_ConvertUnits \
                        "$pvar(width)$pvar(oldunits)" $pvar(units)]
                set pvar(height) [OMF_ConvertUnits \
                        "$pvar(height)$pvar(oldunits)" $pvar(units)]
                set pvar(oldunits) $pvar(units)
            }
        }
        default {}
    }
    # Strip extra precision off of width & height
    if [regexp {\....} $pvar(width)] {
        set pvar(width) [format "%.2f" $pvar(width)]
    }
    if [regexp {\....} $pvar(height)] {
        set pvar(height) [format "%.2f" $pvar(height)]
    }

    # Re-enable trace
    trace variable pvar w OMF_PrintDialogTrace
}


proc OMF_PrintDialogOK { window parrname } {
    OMF_PrintDialogCopyOut $window $parrname
    upvar #0 $parrname parr
    set parr(done) "OK"
    OMF_PrintDialogKill $window
    # Note: The outside trace on parr(done) is removed when the
    # PrintDialog window is destroyed, so it is necessary to write
    # "OK" to parr(done) *before* calling OMF_PrintDialogKill.
}

proc OMF_PrintDialogCancel { window parrname } {
    upvar #0 $parrname parr
    set parr(done) "Cancel"
    OMF_PrintDialogKill $window
}

# Here $parrname must be a global variable
proc OMF_PrintDialog { window title parrname } {
    toplevel $window
    wm group $window .
    wm title $window "$title"
    OMF_PositionChild $window  ;# Position at +.25+.25 over '.'
    upvar #0 $parrname parr
    # parr is a *global* "Printer Array" structure:
    #    canvas     -   name of canvas widget
    #    filename   -   name of output file or pipe
    #    orient     -   portrait | landscape
    #    size       -   output width *or* height (not both), in the format
    #                     '-pagewidth 3.i'  (3 inches width), or
    #                     '-pageheight 5.c' (5 cm high), for example
    #    papertype  -   letter | legal | A4 | A3
    #    crop       -    1 | 0  (1=>print only displayed region of canvas)
    #    xmargin    -    left/right margin, with units.  E.g.: 1.i
    #    ymargin    -    top/bottom margin,  as above
    #    position   -    One of: c nw n ne e se s sw w

    # Packing frames
    set col1frame [frame $window.row1frame -bd 0]
    set col2frame [frame $window.row2frame -bd 0]
    set outputwidgets [frame $window.outputwidgets -relief ridge -bd 4]
    set sizewidgets   [frame $window.sizewidgets   -relief ridge -bd 4]
    set actionwidgets [frame $window.actionwidgets -relief ridge -bd 4]

    # Use $printvar() for "local" global variables
    OMF_GetID $window id
    set printvar "OMFpd$id"
    global $printvar
    OMF_PrintDialogCopyIn $window $parrname ;# Set initial values

    # Filename
    OMF_MakeEntryWindow $window.file "Print to:" {} \
            ${printvar}(filename) 15 0 raised
    bind $window.file.entryvalue <Key-Escape> "set ${printvar}(filename) {}"

    # Orientation
    OMF_MakeRadioList $window.orient 1 0 "Orientation:" {} \
            ${printvar}(orient) {} Portrait Landscape

    # Paper type
    OMF_MakeRadioList $window.papertype 2 0 "Paper type" {} \
            ${printvar}(papertype) {} Letter Legal A4 A3

    # Position
    OMF_MakeRadioList $window.horpos 2 0 "Horizontal Position" {} \
            ${printvar}(horpos) {} {Left w} {Center ""} {Right e}
    OMF_MakeRadioList $window.verpos 2 0 "Vertical Position" {} \
            ${printvar}(verpos) {} {Top n} {Center ""} {Bottom s}

    # Units
    OMF_MakeRadioList $window.units 1 1 "Units:" {} \
            ${printvar}(units) {} {in i} {cm c}

    # Margins
    frame $window.marginframe -relief ridge -bd 2
    OMF_MakeEntryWindow $window.topmargin "Margin Top:" 12 \
            ${printvar}(topmargin) 6 0 flat
    OMF_MakeEntryWindow $window.leftmargin "Left:" 7 \
            ${printvar}(leftmargin) 6 0 flat
    pack $window.topmargin $window.leftmargin \
            -side left -in $window.marginframe

    # Print dimensions
    frame $window.dimframe -relief ridge -bd 2
    OMF_MakeEntryWindow $window.width "Print Width:" 12 \
            ${printvar}(width) 6 0 flat
    OMF_MakeEntryWindow $window.height "Height:" 7 \
            ${printvar}(height) 6 0 flat
    pack $window.width $window.height -side left -in $window.dimframe

    # Crop?
    frame $window.cropframe -relief ridge -bd 2
    checkbutton $window.crop -text "Crop to view"  -relief flat \
            -variable ${printvar}(crop)
    pack $window.crop -side left -in $window.cropframe -expand 1

    # Action buttons
    button $window.ok -text "OK" -width 8 \
            -command "OMF_PrintDialogOK $window $parrname"
    button $window.reset -text "Reset" -width 8 \
            -command "OMF_PrintDialogCopyIn $window $parrname"
    button $window.cancel -text "Cancel" -width 8 \
            -command "OMF_PrintDialogCancel $window $parrname"
    wm protocol $window WM_DELETE_WINDOW "$window.cancel invoke"

    # Pack
    pack $window.file $window.orient $window.papertype \
            $window.horpos $window.verpos \
            -side top -fill x -expand 1 -in $outputwidgets 
    pack $window.units $window.marginframe $window.dimframe \
            $window.cropframe \
            -side top -fill x -expand 1 -in $sizewidgets
    pack $window.ok $window.reset $window.cancel -side left -expand 1 \
            -in $actionwidgets

    pack $outputwidgets -side top -fill x -in $col1frame \
            -padx 5 -pady 5 -expand 1 -anchor n
    pack $sizewidgets   -side top -fill x -in $col2frame \
            -padx 5 -pady 5 -anchor n
    pack $actionwidgets -side top -in $col2frame -fill x \
             -padx 5 -pady 5 -expand 1 -anchor s
    pack $col1frame -side left -fill both -expand 1
    pack $col2frame -side left -fill y

    # Trace to keep printvar internally consistent
    trace variable $printvar w OMF_PrintDialogTrace

    # Trace to update printvar if viewport in . changes
    # This is a little tricky, because we need to send the
    # name $printvar through the trace (which is not tied
    # to $printvar, but to global variable UpdateViewCount.
    # There is probably a cleaner way to do this...  -mjd 97/2/11
    proc OMF_PrintDialogViewportTrace$id { name elt op } \
            "OMF_PrintDialogUpdateVariables $printvar"
    global UpdateViewCount
    trace variable UpdateViewCount w OMF_PrintDialogViewportTrace$id

    MJD_SetIcon $window
    focus $window.file.entryvalue
}

###################### OMF_FileDialog ######################################
# The following global variables are created when OMF_FileDialog is
# called.  The suffix $id (where $id is derived from the name of the
# FileDialog toplevel window---passed in the proc call---by replacing
# each '.' with a '_',) serves to make each collection unique to the
# current dialog.  This way, multiple FileDialog's can be active at
# one time under the same TCL interpreter (provided they have different
# toplevel window names).
#
# OMF_FDPathSelect$id        ""     ;# Path selection window value
# OMF_FDFileSelect$id        ""     ;# File selection window value
# OMF_FDFileFilter$id        "*"    ;# Current filename filter
# OMF_FDFileFilter${id}Entry ""     ;# Filename filter window value
# OMF_FDOrigDir$id           ""     ;# Cwd on FileDialog invocation
# OMF_FDDirCur$id            ""     ;# Current working directory
# OMF_FDDirCur${id}Entry     ""     ;# Cwd window value
# OMF_FDBrowseState$id       0      ;# Close dialog on 'OK' iff 0
# OMF_FDFilterExport$id      ""     ;# Dialog export, file filter
# OMF_FDPathExport$id        ""     ;# Dialog export, path
# OMF_FDFileExport$id        ""     ;# Dialog export, filename(s)
#
# Calling routines can set a trace on OMF_FD*Export$id to detect
# action from FileDialog.  The OMF_FDFileExport variable is filled
# with a space seperated list of files, all in directory
# OMF_FDPathExport.  These variables are guaranteed to be updated
# together, in the order FilterExport, PathExport, FileExport, so
# if a trace is desired on this group then it should be placed on
# FileExport (so that when the trace is executed all 3 variables
# are current).
#
# Note:     The idiom "regexp {^(.*)\.wFD(.*)$} $w full base ext"
#        is used in binding proc's to extract the toplevel dialog
#        window name (base).  When a binding is activated, the %W
#        parameter gives the name of the bound widget.  In the
#        hierarchy used in the OMF_FileDialog scripts, all widgets
#        have a name with the form, for example, of <base>.wFDdir or
#        <base>.wFDdircurframe.value, etc.  So the aforementioned
#        regexp allows a proc to easily extract the <base> and
#        <ext> portions of the widget name.  (The way regexp's are
#        matched, if $w contains more than one instance of .wFD,
#        then the split will occur on the rightmost instance.)
#        Note: Tk doesn't allow widget identifiers to begin with
#        uppercase letters, which is why we use .wFD* instead of
#        just .FD*
#           The top level widget name is used to produce a call-
#        specific identifier code used to name all-specific global
#        variables.  The id is produced by replacing all .'s with
#         _'s, via
#                      regsub -all {\.} "$base" _ id
#        This is convenient because the TCL parser wants to interpret
#        an expression like "$silly.bad.boy" as "${silly}.bad.boy".
#        (One can force the issue with ${silly.bad.boy}, but that is
#        rather awkward.) Note: The TCL parser also doesn't like hyphens
#        in variable names.
#           Then use, e.g.,
#                      global OMF_FDDirCur$id
#        if a lval is needed (i.e., variable name), or
#                      upvar #0 OMF_FDDirCur$id dircur
#        if a rval is needed (i.e., variable value).  Now, of course,
#        the second one also yields a valid lval as well, but (IMPORTANT
#        NOTE!!!) if that variable is bound by a trace, the trace proc
#        will receive the local name (here dircur) instead of the global
#        name (here OMF_FDDirCur$id), and trace proc's like OMF_EntryFollow
#        that depend on the global name (e.g., to build the name of an
#        associated global variable) will break badly, with REALLY OBSCURE
#        ERROR MESSAGES!!!
#
# Directory notes:  There is some strangeness with directory handling.  If
#        the "Cancel" of "Done" button is pressed to exit, then the dialog
#        does a 'cd' to the current directory at the time the dialog was
#        initially called.  Also, the dialog does a 'cd' anytime the user
#        changes the path (either in the path entry box or by double clicking
#        in the directory listbox).  The dialog also does a 'cd' to what it
#        is storing as the current directory (as shown in its path entry box)
#        each time it updates its filelist, to guard against other proc's
#        changing of cwd (current working directory).  As a result, multiple
#        dialog boxes should work properly simultaneously, but the cwd seen
#        by the calling application (and dialogs opened in the future) can be
#        difficult to predict.  Perhaps in the future, I'll rewrite things
#        so that these dialog's only temporarily change the cwd, and allow
#        the calling routine to control permanent change to the cwd as it
#        deems fit.
#

proc OMF_FDDirSelect { w } {
    # Used by directory listbox bindings
    regexp {^(.*)\.wFD(.*)$} $w full base ext
    regsub -all {\.} "$base" _ id
    global OMF_FDDirCur$id
    upvar #0 OMF_FDDirCur$id dircur
    upvar #0 OMF_FDFileFilter$id filter
    set index [$w  curselection]
    if { "$index" == "" } then return
    set temp "${dircur}/[$w get $index]"
    if [file isdirectory "$temp"] then {
        cd $temp
        set OMF_FDDirCur$id [pwd]
        OMF_FDFillFileLists $base $base.wFDdir $base.wFDfile "$filter"
    }
}

proc OMF_FDClearSelection { base } {
    regsub -all {\.} "$base" _ id
    global OMF_FDPathSelect$id OMF_FDFileSelect$id
    set OMF_FDPathSelect$id ""
    set OMF_FDFileSelect$id ""
}

proc OMF_FDSelect { w } {
    # Used by file listbox bindings
    regexp {^(.*)\.wFD(.*)$} $w full base ext
    regsub -all {\.} "$base" _ id
    upvar #0 OMF_FDDirCur$id curdir
    global OMF_FDPathSelect$id
    global OMF_FDFileSelect$id
    set selectindexlist [$w  curselection]
    set selectlist {}
    foreach i $selectindexlist {
        lappend selectlist [$w get $i]
    }
    set OMF_FDPathSelect$id $curdir
    set OMF_FDFileSelect$id $selectlist
}


proc OMF_FDFillFileLists { base dirlist filelist filefilter } {
    regsub -all {\.} "$base" _ id

    OMF_PushCursor
    OMF_SoftUpdate

    upvar #0 "OMF_FDFileFilter$id" ffg
    global OMF_FDFileFilter$id
    set OMF_FDFileFilter$id "$filefilter"
    upvar #0 OMF_FDDirCur$id dircur
    cd $dircur        ;# Just in case some other "dialog" changed the cwd
    $filelist delete 0 end
    $dirlist delete 0 end
    $dirlist insert end ".."
    foreach i [lsort [glob -nocomplain *]] {
        if { [file isfile $i] } then {
	    foreach j $filefilter {
		if [string match $j $i] then {
		    $filelist insert end $i
		    break
		}
	    }
        } elseif { [file isdirectory $i] } then {
            $dirlist insert end $i
        }
    }

    OMF_PopCursor
}

# Rescans directory and fills file & directory lists
proc OMF_FDRescan { base } {
    regsub -all {\.} "$base" _ id
    upvar #0 OMF_FDFileFilter$id filefilter
    OMF_FDFillFileLists $base $base.wFDdir $base.wFDfile "$filefilter"
}

proc OMF_FDInit { w filter_init } {
    # Creates & inits call-specific global variables & traces
    regsub -all {\.} "$w" _ id
    global OMF_FDPathSelect$id        ;# Path selection window value
    global OMF_FDFileSelect$id        ;# File selection window value
    global OMF_FDFileFilter$id        ;# Current filename filter
    global OMF_FDFileFilter${id}Entry ;# Filename filter window value
    global OMF_FDOrigDir$id           ;# Cwd on FileDialog invocation
    global OMF_FDDirCur$id            ;# Current working directory
    global OMF_FDDirCur${id}Entry     ;# Cwd window value
    global OMF_FDBrowseState$id       ;# Close dialog on 'OK' iff 0
    global OMF_FDFilterExport$id      ;# Dialog export, file filter
    global OMF_FDPathExport$id        ;# Dialog export, path
    global OMF_FDFileExport$id        ;# Dialog export, filename 
    trace variable OMF_FDFileFilter$id w OMF_EntryFollow
    trace variable OMF_FDDirCur$id w OMF_EntryFollow

    set OMF_FDFileFilter${id}Entry ""
    set OMF_FDDirCur${id}Entry     ""
    # Note: *Entry variables get initialized through traces, so just
    #       make sure they exist before their shadow-ered variables
    #       are set.
    set OMF_FDPathSelect$id      ""
    set OMF_FDFileSelect$id      ""
    set OMF_FDFileFilter$id      "$filter_init"
    set OMF_FDOrigDir$id         "[pwd]"
    set OMF_FDDirCur$id          "[pwd]"
    set OMF_FDBrowseState$id     0
    set OMF_FDFilterExport$id    "*"
    set OMF_FDPathExport$id      ""
    set OMF_FDFileExport$id      ""
}

proc OMF_FDDestroy { w } {
    # Destroys widgets and unsets call-specific global variables & traces
    regsub -all {\.} "$w" _ id
    global OMF_FDPathSelect$id        ;# Path selection window value
    global OMF_FDFileSelect$id        ;# File selection window value
    global OMF_FDFileFilter$id        ;# Current filename filter
    global OMF_FDFileFilter${id}Entry ;# Filename filter window value
    global OMF_FDOrigDir$id           ;# Cwd on FileDialog invocation
    global OMF_FDDirCur$id            ;# Current working directory
    global OMF_FDDirCur${id}Entry     ;# Cwd window value
    global OMF_FDBrowseState$id       ;# Close dialog on 'OK' iff 0
    trace vdelete OMF_FDFileFilter$id w OMF_EntryFollow
    trace vdelete OMF_FDDirCur$id w OMF_EntryFollow
    unset OMF_FDPathSelect$id
    unset OMF_FDFileSelect$id
    unset OMF_FDFileFilter$id
    unset OMF_FDFileFilter${id}Entry
    unset OMF_FDOrigDir$id
    unset OMF_FDDirCur$id
    unset OMF_FDDirCur${id}Entry
    unset OMF_FDBrowseState$id
    # Note: The OMF_FD*Export$id variables can only be unset by
    #       the calling routine, since these variables are used
    #       to pass back results.
    OMF_SetFocus . $w ;# Make sure . has focus before we destroy $w
    destroy $w
}

# Command invoked by FileDialog "OK" button.
# This is written as an independent procedure to enable runtime
# resolution of "base"
proc OMF_FDokcmd { base } {
    regsub -all {\.} "$base" _ id
    global OMF_FDFilterExport$id OMF_FDPathExport$id OMF_FDFileExport$id
    global OMF_FDFileFilter$id OMF_FDPathSelect$id OMF_FDFileSelect$id
    upvar #0 OMF_FDFileFilter$id filter
    upvar #0 OMF_FDPathSelect$id path
    upvar #0 OMF_FDFileSelect$id file
    # We need to 'set' to the global name, or else
    # traces won't work properly.  (Not that we care...)
    set OMF_FDFileFilter$id [string trim $filter]
    set OMF_FDPathSelect$id [string trim $path]
    set OMF_FDFileSelect$id [string trim $file]
    if {[string compare "$file" ""]==0} { 
        OMF_FDcancelcmd $base
    } else {
        set OMF_FDFilterExport$id "$filter"
        set OMF_FDPathExport$id   "$path"
        set OMF_FDFileExport$id   "$file"
        upvar #0 OMF_FDBrowseState$id browsestate
        if { !$browsestate } then {
            OMF_FDDestroy $base
        }
    }
}

# Command invoked by FileDialog "Cancel" button.
# This is written as an independent procedure to enable runtime
# resolution of "base"
proc OMF_FDcancelcmd { base } {
    regsub -all {\.} "$base" _ id
    upvar #0 OMF_FDOrigDir$id origdir
    cd $origdir
    OMF_FDDestroy $base
}

# Command invoked by FileDialog "Browse" checkbox.
# This is written as an independent procedure to enable runtime
# resolution of "base"
proc OMF_FDbrowsecmd { base } {
    regsub -all {\.} "$base" _ id
    upvar #0 OMF_FDBrowseState$id browsestate
    if { $browsestate } then {
        $base.wFDcancelbtn configure -text "Dismiss"
    } else {
        $base.wFDcancelbtn configure -text "Cancel"
    }
}


### proc OMF_FileDialog ################
proc OMF_FileDialog { w title fileselectmode allow_browse default_filter } {
    # Force fileselectmode to be one of "browse" or "extended"
    # "single" and "multi" are also valid listbox modes, and
    # this code might be changed to accept them in the future.
    # -mjd 20-Jan-1997
    if { ![string compare "$fileselectmode" "multi"] || \
            ![string compare "$fileselectmode" "extended"] } then {
        set fileselectmode "extended"
    } else {
        set fileselectmode "browse"
    }
    if { "$allow_browse" != "0" && "$allow_browse" != "" } then {
        set allow_browse 1
    } else {
        set allow_browse 0
    }
    if { "$default_filter" == "" } then {
        set default_filter "*"
    }

    OMF_FDInit $w $default_filter

    regsub -all {\.} "$w" _ id     ;# Contruct ID from window name
    global OMF_FDPathSelect$id        ;# Path selection window value
    global OMF_FDFileSelect$id        ;# File selection window value
    global OMF_FDFileFilter$id        ;# Current filename filter
    global OMF_FDFileFilter${id}Entry ;# Filename filter window value
    global OMF_FDOrigDir$id           ;# Cwd on FileDialog invocation
    global OMF_FDDirCur$id            ;# Current working directory
    global OMF_FDDirCur${id}Entry     ;# Cwd window value
    global OMF_FDBrowseState$id       ;# Close dialog on 'OK' iff 0

    global omfTroughColor

    toplevel $w
    # Note: WM_DELETE_WINDOW wm protocol is tied below is tied to
    # the "wFDcancelbtn invoke" action.
    wm group $w .
    wm title $w "$title"
    OMF_PositionChild $w  ;# Position at +.25+.25 over '.'

    # Directory widgets #################
    frame $w.wFDdirframe -relief ridge -bd 4 ;# -bg blue

    OMF_MakeEntryWindow $w.wFDdircurframe "Path:" {} \
            OMF_FDDirCur${id}Entry 30 0
    # Note: It is possible to use a command like
    #       $w.wFDdircurframe.entryvalue xview end
    #   to display the rightmost half of the path, but this can
    #   be misleading if the leftmost visible character is a '/',
    #   i.e., the user may think he is seeing the entire path and
    #   believe he is off a directory closer to root (i.e., '/')
    #   than he really is.
    bind $w.wFDdircurframe.entryvalue <Key-Return> {
        regexp {^(.*)\.wFD(.*)$} %W full base ext
        regsub -all {\.} "$base" _ id
        global OMF_FDDirCur$id OMF_FDDirCur${id}Entry
        upvar #0 OMF_FDDirCur$id dircur
        # See earlier note about using upvar'ed variables as lval's
        # of shadowed variables (i.e., trace'd through OMF_EntryFollow)
        upvar #0 OMF_FDDirCur${id}Entry dircurentry
        if [file isdirectory "$dircurentry"] then {
            cd "$dircurentry"
            set OMF_FDDirCur$id [pwd]
            OMF_FDRescan $base
        } else {
            set OMF_FDDirCur${id}Entry "$dircur"
            # For this one we should be able to use name dircurentry
            # for lval, but why take chances? ;-)
        }
    }
    bind $w.wFDdircurframe.entryvalue <Key-Escape> {
        regexp {^(.*)\.wFD(.*)$} %W full base ext
        regsub -all {\.} "$base" _ id
        upvar #0 OMF_FDDirCur$id dircur
        set OMF_FDDirCur${id}Entry "$dircur"
    }
    bind $w.wFDdircurframe.entryvalue <FocusOut> {
        # I really want this to just invoke the <Key-Escape>
        # binding so I don't have to duplicate that code, but
        # at the present time (20-Jan-1997) I don't know how
        # to do this.  -mjd
        regexp {^(.*)\.wFD(.*)$} %W full base ext
        regsub -all {\.} "$base" _ id
        upvar #0 OMF_FDDirCur$id dircur
        set OMF_FDDirCur${id}Entry "$dircur"
    }

    frame $w.wFDdirlistboxframe  -relief raised -bd 2
    listbox $w.wFDdir -bd 0 -selectmode browse \
            -yscrollcommand "$w.wFDscrdir set" -height 8
    bind $w.wFDdir <ButtonPress> "focus %W"
    bind $w.wFDdir <Double-Button> "OMF_FDDirSelect %W"
    bind $w.wFDdir <Key-Return> "OMF_FDDirSelect %W"
    scrollbar $w.wFDscrdir -bd 2 -troughcolor $omfTroughColor \
            -command "$w.wFDdir yview"
    pack $w.wFDdir -side left -fill both -expand 1 \
            -in $w.wFDdirlistboxframe
    pack $w.wFDscrdir -side right -fill y -in $w.wFDdirlistboxframe

    pack $w.wFDdircurframe \
            -side top -anchor w -fill x -in $w.wFDdirframe
    pack $w.wFDdirlistboxframe \
            -side top -anchor w -fill both -expand 1 -in $w.wFDdirframe


    # File-selection widgets #################
    frame $w.wFDfileframe -relief ridge -bd 4

    OMF_MakeEntryWindow $w.wFDfilefilterframe "Filter:" {} \
            OMF_FDFileFilter${id}Entry 23 0
    bind $w.wFDfilefilterframe.entryvalue <Key-Return> {
        regexp {^(.*)\.wFD(.*)$} %W full base ext
        regsub -all {\.} "$base" _ id
        upvar #0 OMF_FDFileFilter${id}Entry newfilter
        global OMF_FDFileFilter${id}
        set OMF_FDFileFilter${id} "$newfilter"
        OMF_FDRescan $base
    }
    bind $w.wFDfilefilterframe.entryvalue <Key-Escape> {
        regexp {^(.*)\.wFD(.*)$} %W full base ext
        regsub -all {\.} "$base" _ id
        upvar #0 OMF_FDFileFilter$id filefilter
        set OMF_FDFileFilter${id}Entry $filefilter
    }
    bind $w.wFDfilefilterframe.entryvalue <FocusOut> {
        # I really want this to just invoke the <Key-Escape>
        # binding so I don't have to duplicate that code, but
        # at the present time (20-Jan-1997) I don't know how
        # to do this.  -mjd
        regexp {^(.*)\.wFD(.*)$} %W full base ext
        regsub -all {\.} "$base" _ id
        upvar #0 OMF_FDFileFilter$id filefilter
        set OMF_FDFileFilter${id}Entry $filefilter
    }

    frame $w.wFDfilelistboxframe  -relief raised -bd 2
    listbox $w.wFDfile -bd 0 -selectmode $fileselectmode -height 15 \
            -xscrollcommand "$w.wFDfilescrx set" \
            -yscrollcommand "$w.wFDfilescry set"
    bind $w.wFDfile <ButtonPress> "focus %W"
    bind $w.wFDfile <ButtonRelease> "OMF_FDSelect %W"
    bind $w.wFDfile <Key-Return> \
            "OMF_FDSelect $w.wFDfile ; $w.wFDokbtn invoke" 

    bind $w.wFDfile <Double-Button> "$w.wFDokbtn invoke; break"
    # Be sure to "break" out of the binding, so we don't
    # process any other (in particular, class) bindings
    # after the widget is destroyed! (-mjd, 28-Jan-1997)
    scrollbar $w.wFDfilescry -bd 2 -troughcolor $omfTroughColor \
            -command "$w.wFDfile yview"
    scrollbar $w.wFDfilescrx -bd 2 -troughcolor $omfTroughColor \
            -orient horizontal -command "$w.wFDfile xview"
    pack $w.wFDfilescry -side right -fill y -in $w.wFDfilelistboxframe
    pack $w.wFDfilescrx -side bottom -fill x -in $w.wFDfilelistboxframe
    pack $w.wFDfile -side top -fill both -expand 1 \
            -in $w.wFDfilelistboxframe

    pack $w.wFDfilefilterframe \
            -side top -anchor w -fill x -in $w.wFDfileframe
    pack $w.wFDfilelistboxframe \
            -side top -anchor w -fill both -expand 1 -in $w.wFDfileframe

    # Selection widgets #################
    frame $w.wFDselectframe -relief ridge -bd 4
    label $w.wFDselectframe.label -text "Selection" \
            -relief raised -bd 2
    OMF_MakeEntryWindow $w.wFDselectframe.pathframe "Path:" 5 \
            OMF_FDPathSelect$id 20 0
    if { ![string compare "$fileselectmode" "multi"] || \
            ![string compare "$fileselectmode" "extended"] } then {
        OMF_MakeEntryWindow $w.wFDselectframe.fileframe "Files:" 5 \
                OMF_FDFileSelect$id 20 1
    } else {
        OMF_MakeEntryWindow $w.wFDselectframe.fileframe "File:" 5 \
                OMF_FDFileSelect$id 20 0
    }
    bind $w.wFDselectframe.pathframe.entryvalue \
            <Key-Return> "$w.wFDokbtn invoke" 
    bind $w.wFDselectframe.fileframe.entryvalue \
            <Key-Return> "$w.wFDokbtn invoke"
    pack $w.wFDselectframe.label -side top -fill x
    pack $w.wFDselectframe.pathframe $w.wFDselectframe.fileframe \
            -side top -fill x -expand 1

    # Operation ("OK", "Cancel", etc.) widgets #################
    frame $w.wFDopframe -relief ridge -bd 4
    frame $w.wFDopframecol1 -relief flat -bd 0
    frame $w.wFDopframecol2 -relief flat -bd 0
    button $w.wFDokbtn -text "OK" -command "OMF_FDokcmd $w"
    # Caution: Double-clicking in file selection box, or hitting
    # <Key-Return> in either of the selection entry widgets also
    # invokes the $w.wFDokbtn, so if this name or action is changed,
    # update those bindings as necessary.
    button $w.wFDcancelbtn -text "Cancel" -width 6 \
            -command "OMF_FDcancelcmd $w"
    wm protocol $w WM_DELETE_WINDOW "$w.wFDcancelbtn invoke"
    button $w.wFDrescanbtn -text "Rescan" -command "OMF_FDRescan $w"
    checkbutton $w.wFDbrowsebtn -text Browse -relief raised \
            -variable OMF_FDBrowseState$id
    if { $allow_browse } then {
        $w.wFDbrowsebtn configure -command "OMF_FDbrowsecmd $w"
    } else {
        $w.wFDbrowsebtn configure -command "" -state disabled
    }
    button $w.wFDclearbtn -text "Clear" -relief raised \
            -command "OMF_FDClearSelection $w
                      set OMF_FDFileFilter$id \"*\"
                      OMF_FDRescan $w"

    pack $w.wFDokbtn $w.wFDcancelbtn -side top \
            -fill both -expand 1 -in $w.wFDopframecol2
    pack $w.wFDbrowsebtn $w.wFDclearbtn $w.wFDrescanbtn -side top \
            -fill both -expand 1 -in $w.wFDopframecol1
    pack $w.wFDopframecol1 $w.wFDopframecol2 \
            -side left -fill both -expand 1 -in $w.wFDopframe

    pack $w.wFDfileframe -side left -fill y -padx 2m -pady 2m
    pack $w.wFDopframe -side top -anchor e -padx 2m -pady 2m
    pack $w.wFDselectframe -side top -fill x -anchor w -padx 2m -pady 2m
    pack $w.wFDdirframe -side bottom -fill both -expand 1 \
            -anchor w -padx 2m -pady 2m

    # Adjust tabbing order
    # ...File list widgets
    raise $w.wFDfilefilterframe
    raise $w.wFDfile
    raise $w.wFDfilescry
    raise $w.wFDfilescrx
    # ...Command button widgets
    raise $w.wFDokbtn
    raise $w.wFDcancelbtn
    raise $w.wFDbrowsebtn
    raise $w.wFDclearbtn
    raise $w.wFDrescanbtn
    # ...Selection widgets
    raise $w.wFDselectframe
    raise $w.wFDselectframe.pathframe
    raise $w.wFDselectframe.fileframe
    # ...Directory list widgets
    raise $w.wFDdircurframe
    raise $w.wFDdir
    raise $w.wFDscrdir

    # Read filenames & insert into listboxes
    OMF_FDRescan $w

    MJD_SetIcon $w
    focus $w.wFDfile
}

